package table

import error.Severity
import function.FunctionNames
import function.LispFunction
import org.assertj.core.api.Assertions.assertThat
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Assertions.assertThrows
import org.junit.jupiter.api.BeforeEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.TestInstance
import org.junit.jupiter.api.TestInstance.Lifecycle.PER_CLASS
import sexpression.Cons
import sexpression.Nil
import sexpression.Nil.NIL
import sexpression.Symbol.T
import table.FunctionTable.LispFunctionInstantiationException
import table.FunctionTable.defineFunction
import table.FunctionTable.isAlreadyDefined
import table.FunctionTable.lookupFunction
import table.FunctionTable.resetFunctionTable
import java.util.HashSet

@TestInstance(PER_CLASS)
class FunctionTableTest {

    @FunctionNames("GOOD")
    class GoodFunction(name: String) : LispFunction() {

        override fun call(argumentList: Cons): Nil = NIL
    }

    @FunctionNames("BAD")
    class BadFunction : LispFunction() {

        override fun call(argumentList: Cons): Nil = NIL
    }

    class UglyFunction : LispFunction() {

        override fun call(argumentList: Cons): Nil = NIL
    }

    private fun createLispFunction() = object : LispFunction() {
        override fun call(argumentList: Cons) = T
    }

    @BeforeEach
    fun initialize() {
        resetFunctionTable()
    }

    @AfterEach
    fun tearDown() {
        resetFunctionTable()
    }

    @Test
    fun builtInFunctionIsDefined() {
        assertThat(isAlreadyDefined("CONS")).isTrue()
    }

    @Test
    fun undefinedFunctionIsNotDefined() {
        assertThat(isAlreadyDefined("undefined")).isFalse()
    }

    @Test
    fun lookupBuiltInFunction_ReturnsFunction() {
        assertThat(lookupFunction("CONS")).isNotNull()
    }

    @Test
    fun lookupUndefinedFunction_ReturnsNull() {
        assertThat(lookupFunction("undefined")).isNull()
    }

    @Test
    fun defineFunctionWorks() {
        val functionName = "testFunction"
        val testFunction = createLispFunction()

        assertThat(lookupFunction(functionName)).isNull()
        assertThat(isAlreadyDefined(functionName)).isFalse()

        defineFunction(functionName, testFunction)

        assertThat(isAlreadyDefined(functionName)).isTrue()
        assertThat(lookupFunction(functionName)).isEqualTo(testFunction)
    }

    @Test
    fun resetFunctionTableWorks() {
        val functionName = "testFunction"
        val testFunction = createLispFunction()
        defineFunction(functionName, testFunction)

        resetFunctionTable()

        assertThat(lookupFunction(functionName)).isNull()
        assertThat(isAlreadyDefined(functionName)).isFalse()
    }

    @Test
    fun resetWithCustomBuitIns() {
        val goodBuiltIns = HashSet<Class<out LispFunction>>()
        goodBuiltIns.add(GoodFunction::class.java)

        resetFunctionTable(goodBuiltIns)

        assertThat(isAlreadyDefined("GOOD")).isTrue()
        assertThat(lookupFunction("GOOD")).isNotNull()
    }

    @Test
    fun unableToInitializeBuiltIn() {
        val badBuiltIns = HashSet<Class<out LispFunction>>()
        badBuiltIns.add(BadFunction::class.java)

        assertThrows(LispFunctionInstantiationException::class.java) { resetFunctionTable(badBuiltIns) }
    }

    @Test
    fun lispFunctionInstantiationException_HasCorrectAttributes() {
        val e = LispFunctionInstantiationException("Bad")

        assertThat(e.message).isNotEmpty()
        assertThat(e.severity).isEqualTo(Severity.CRITICAL)
    }

    @Test
    fun namelessBuiltIn_DoesNotCauseNPE() {
        val namelessBuiltins = HashSet<Class<out LispFunction>>()
        namelessBuiltins.add(UglyFunction::class.java)

        resetFunctionTable(namelessBuiltins)

        assertThat(isAlreadyDefined("UGLY")).isFalse()
        assertThat(lookupFunction("UGLY")).isNull()
    }
}
